/*********************************************************************
 *
 *	Transmission Control Protocol (TCP) Communications Layer
 *  Module for Microchip TCP/IP Stack
 *	 -Provides reliable, handshaked transport of application stream 
 *    oriented data with flow control
 *	 -Reference: RFC 793
 *
 *********************************************************************
 * FileName:        TCP.c
 * Dependencies:    string.h
 *                  StackTsk.h
 *                  Helpers.h
 *                  IP.h
 *                  MAC.h
 *                  ARP.h
 *                  Tick.h
 *                  TCP.h
 * Processor:       PIC18, PIC24F, PIC24H, dsPIC30F, dsPIC33F
 * Compiler:        Microchip C18 v3.02 or higher
 *					Microchip C30 v2.01 or higher
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * Copyright  2002-2007 Microchip Technology Inc.  All rights 
 * reserved.
 *
 * Microchip licenses to you the right to use, modify, copy, and 
 * distribute: 
 * (i)  the Software when embedded on a Microchip microcontroller or 
 *      digital signal controller product (Device) which is 
 *      integrated into Licensees product; or
 * (ii) ONLY the Software driver source files ENC28J60.c and 
 *      ENC28J60.h ported to a non-Microchip device used in 
 *      conjunction with a Microchip ethernet controller for the 
 *      sole purpose of interfacing with the ethernet controller. 
 *
 * You should refer to the license agreement accompanying this 
 * Software for additional information regarding your rights and 
 * obligations.
 *
 * THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT 
 * WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT 
 * LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A 
 * PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL 
 * MICROCHIP BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR 
 * CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF 
 * PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS 
 * BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE 
 * THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER 
 * SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT 
 * (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.
 *
 *
 * Author               Date    	Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Nilesh Rajbharti     5/8/01  	Original        (Rev 1.0)
 * Howard Schlunder		12/11/06	Changed almost everything to 
 *									better meet RFC 793.
 ********************************************************************/
#define __TCP_C

#include "TCPIP Stack/TCPIP.h"

#if defined(STACK_USE_TCP)

// Configurable TCP options
#define TCP_MAX_SEG_SIZE			(536)

// TCP Timeout and retransmit numbers
#define TCP_START_TIMEOUT_VAL   	((TICK)TICK_SECOND*1)
#define TCP_DELAYED_ACK_TIMEOUT		((TICK)TICK_SECOND/10)
#define TCP_FIN_WAIT_2_TIMEOUT		((TICK)TICK_SECOND*5)
#define TCP_MAX_RETRIES			    (5u)
#define TCP_MAX_SYN_RETRIES			(2u)	// Smaller than all other retries to reduce SYN flood DoS duration

#define TCP_AUTO_TRANSMIT_TIMEOUT_VAL	(TICK_SECOND/25ull)

// TCP Flags defined in RFC
#define FIN     (0x01)
#define SYN     (0x02)
#define RST     (0x04)
#define PSH     (0x08)
#define ACK     (0x10)
#define URG     (0x20)

// TCP Header
typedef struct _TCP_HEADER
{
	WORD    SourcePort;
	WORD    DestPort;
	DWORD   SeqNumber;
	DWORD   AckNumber;

	struct
	{
		unsigned char Reserved3      : 4;
		unsigned char Val            : 4;
	} DataOffset;

	union
	{
		struct
		{
			unsigned char flagFIN    : 1;
			unsigned char flagSYN    : 1;
			unsigned char flagRST    : 1;
			unsigned char flagPSH    : 1;
			unsigned char flagACK    : 1;
			unsigned char flagURG    : 1;
			unsigned char Reserved2  : 2;
		} bits;
		BYTE byte;
	} Flags;

	WORD    Window;
	WORD    Checksum;
	WORD    UrgentPointer;
} TCP_HEADER;

// TCP Options as defined by RFC
#define TCP_OPTIONS_END_OF_LIST     (0x00u)
#define TCP_OPTIONS_NO_OP           (0x01u)
#define TCP_OPTIONS_MAX_SEG_SIZE    (0x02u)
typedef struct _TCP_OPTIONS
{
	BYTE        Kind;
	BYTE        Length;
	WORD_VAL    MaxSegSize;
} TCP_OPTIONS;

#define SwapPseudoTCPHeader(h)  (h.TCPLength = swaps(h.TCPLength))

// IP pseudo header as defined by RFC 793
typedef struct _PSEUDO_HEADER
{
	IP_ADDR SourceAddress;
	IP_ADDR DestAddress;
	BYTE Zero;
	BYTE Protocol;
	WORD TCPLength;
} PSEUDO_HEADER;

#define LOCAL_PORT_START_NUMBER (1024u)
#define LOCAL_PORT_END_NUMBER   (5000u)

// For debugging only.  Normal applications should never enable these
//#define DEBUG_GENERATE_TX_LOSS		31500
//#define DEBUG_GENERATE_RX_LOSS		31500


// Local temp port numbers.
#ifdef STACK_CLIENT_MODE
static WORD NextPort __attribute__((persistent));
#endif

#if defined(HI_TECH_C)
	// The initializer forces this large array out of the bss section 
	// so we can link correctly.
	TCB_STUB TCBStubs[MAX_TCP_SOCKETS] = {'\0'};	
#else
	// The TCB array is very large.  With the C18 compiler, one must 
	// modify the linker script to make an array that spans more than 
	// one memory bank.  To do this, make the necessary changes to your 
	// processor's linker script (.lkr).  Here is an example showing 
	// gpr11 and 128 bytes of gpr12 being combined into one 384 byte 
	// block used exclusively by the TCB_MEM data section:
	// ...
	// //DATABANK   NAME=gpr11      START=0xB00          END=0xBFF
	// //DATABANK   NAME=gpr12      START=0xC00          END=0xCFF
	// DATABANK   NAME=gpr11b     START=0xB00          END=0xC7F           PROTECTED
	// DATABANK   NAME=gpr12      START=0xC80          END=0xCFF
	// ...
	// SECTION    NAME=TCP_TCB_RAM    RAM=gpr11b
	// ...
	#pragma udata TCB_uRAM
	TCB_STUB TCBStubs[MAX_TCP_SOCKETS];
	#pragma udata					// Return to any other RAM section
#endif

static TCB MyTCB;


static void HandleTCPSeg(TCP_SOCKET hTCP, TCP_HEADER *h, WORD len);
static void SendTCP(TCP_SOCKET hTCP, BYTE flags, BOOL bResetTimer);
static TCP_SOCKET FindMatchingSocket(TCP_HEADER *h, NODE_INFO *remote);
static void SwapTCPHeader(TCP_HEADER* header);
static void CloseSocket(TCP_SOCKET hTCP);
static void SyncTCB(TCP_SOCKET hTCP);



// Flushes MyTCB cache and loads up the specified TCB
// Does nothing on cache hit
static void SyncTCB(TCP_SOCKET hTCP)
{
	static TCP_SOCKET hCurrentTCB = INVALID_SOCKET;
	WORD PtrSave;
	
	if(hCurrentTCB == hTCP)
		return;

	if(hCurrentTCB != INVALID_SOCKET)
	{
		// Save the current TCB
		PtrSave = MACSetWritePtr(TCBStubs[hCurrentTCB].bufferTxStart - sizeof(MyTCB));
		MACPutArray((BYTE*)&MyTCB, sizeof(MyTCB));
		MACSetWritePtr(PtrSave);
	}

	// Load up the new TCB
	hCurrentTCB = hTCP;
	PtrSave = MACSetReadPtr(TCBStubs[hTCP].bufferTxStart - sizeof(MyTCB));
	MACGetArray((BYTE*)&MyTCB, sizeof(MyTCB));
	MACSetReadPtr(PtrSave);
}


/*********************************************************************
* Function:        void TCPInit(void)
*
* PreCondition:    None
*
* Input:           None
*
* Output:          TCP is initialized.
*
* Side Effects:    None
*
* Overview:        Initialize all socket states
*
* Note:            This function is called only once during lifetime
*                  of the application.
********************************************************************/
void TCPInit(void)
{
	TCP_SOCKET hTCP;
	TCB_STUB *ps;
	
	// Initialize all sockets.
	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++)
	{
		ps = &TCBStubs[hTCP];

		ps->smState			= TCP_CLOSED;
		ps->bufferTxStart	= BASE_TCB_ADDR + sizeof(TCB) + hTCP*(sizeof(TCB) + (TCP_TX_FIFO_SIZE+1) + (TCP_RX_FIFO_SIZE+1));
		ps->bufferRxStart	= ps->bufferTxStart + TCP_TX_FIFO_SIZE+1;
		ps->bufferEnd		= ps->bufferRxStart + TCP_RX_FIFO_SIZE;
		ps->Flags.bServer	= FALSE;
		ps->Flags.bMemoryExternal = FALSE;
		CloseSocket(hTCP);
	}
}



/*********************************************************************
* Function:        TCP_SOCKET TCPListen(WORD port)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           port    - A TCP port to be opened.
*
* Output:          Given port is opened and returned on success
*                  INVALID_SOCKET if no more sockets left.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
TCP_SOCKET TCPListen(WORD port)
{
	TCP_SOCKET hTCP;
	TCB_STUB *ps;

	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++)
	{
		ps = &TCBStubs[hTCP];

		if(ps->smState == TCP_CLOSED)
		{
			ps->Flags.bServer = TRUE;
			ps->smState = TCP_LISTEN;
			ps->remoteHash.Val = port;

			SyncTCB(hTCP);
			MyTCB.localPort.Val = port;
			
			return hTCP;
		}
	}

	return INVALID_SOCKET;
}


/*********************************************************************
* FunctionL		BOOL TCPGetRemoteInfo(TCP_SOCKET hTCP, SOCKET_INFO *RemoteInfo)
*
* PreCondition:	TCPInit() is already called.
*
* Input:		hTCP      - Handle of socket to read
*
* Output:		A new socket is created, connection request is
*				sent and socket handle is returned.
*
* Side Effects:	None
*
* Overview:		None
*
* Note:			None
*
********************************************************************/
SOCKET_INFO* TCPGetRemoteInfo(TCP_SOCKET hTCP)
{
	static SOCKET_INFO	RemoteInfo;

	SyncTCB(hTCP);
	memcpy((void*)&RemoteInfo.remote, (void*)&MyTCB.remote, sizeof(NODE_INFO));
	RemoteInfo.remotePort.Val = MyTCB.remotePort.Val;

	return &RemoteInfo;
}


/*********************************************************************
* Function:        TCP_SOCKET TCPConnect(NODE_INFO* remote,
*                                      WORD remotePort)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           remote      - Remote node address info
*                  remotePort  - remote port to be connected.
*
* Output:          A new socket is created, connection request is
*                  sent and socket handle is returned.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            By default this function is not included in
*                  source.  You must define STACK_CLIENT_MODE to
*                  be able to use this function.
********************************************************************/
#ifdef STACK_CLIENT_MODE
TCP_SOCKET TCPConnect(NODE_INFO *remote, WORD remotePort)
{
	TCP_SOCKET hTCP;
	TCB_STUB *ps;

	// Find an available socket
	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++)
	{
		ps = &TCBStubs[hTCP];

		if(ps->smState != TCP_CLOSED)
			continue;

		SyncTCB(hTCP);

		// Each new socket that is opened by this node, gets the 
		// next sequential port number.
		if(NextPort < LOCAL_PORT_START_NUMBER || NextPort > LOCAL_PORT_END_NUMBER)
			NextPort = LOCAL_PORT_START_NUMBER;
		
		// Set the non-zero TCB fields
		MyTCB.localPort.Val = NextPort++;
		MyTCB.remotePort.Val = remotePort;
		memcpy((void*)&MyTCB.remote, (void*)remote, sizeof(NODE_INFO));

		// Update the TCB Stub
		ps->remoteHash.Val = (remote->IPAddr.w[1]+remote->IPAddr.w[0] + remotePort) ^ MyTCB.localPort.Val;
		ps->smState = TCP_SYN_SENT;
		
		// Send TCP SYNchronize (connect) request
		SendTCP(hTCP, SYN, TRUE);

		return hTCP;
	}

	// If there is no socket available, return error.
	return INVALID_SOCKET;
}
#endif


/*********************************************************************
* Function:        BOOL TCPWasReset(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP    - Socket to be checked
*
* Output:          TRUE: if given socket was reset at one 
*						 point since the last call to 
*						 TCPWasDisconnected().
*                  FALSE: if given socket hasn't been reset since the 
*						  last call to TCPWasReset.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
BOOL TCPWasReset(TCP_SOCKET hTCP)
{
	if(TCBStubs[hTCP].Flags.bSocketReset)
	{
		TCBStubs[hTCP].Flags.bSocketReset = 0;
		return TRUE;
	}	
	
	return FALSE;
}


/*********************************************************************
* Function:        BOOL TCPIsConnected(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP    - Socket to be checked for connection.
*
* Output:          TRUE    if given socket is connected
*                  FALSE   if given socket is not connected.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            A socket is said to be connected only if it is in 
*				   the TCP_ESTABLISHED state
********************************************************************/
BOOL TCPIsConnected(TCP_SOCKET hTCP)
{
	BYTE i = TCBStubs[hTCP].smState;
	return (i == TCP_ESTABLISHED) || (i == TCP_LOOPBACK);
}


/*********************************************************************
* Function:        void TCPDisconnect(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called
*
* Input:           hTCP - Socket to be disconnected.
*
* Output:          A disconnect request is sent for given socket.  
*				   This function does nothing if the socket isn't 
*				   currently connected.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
void TCPDisconnect(TCP_SOCKET hTCP)
{
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];

	// Delete all data in the RX FIFO
	// In this stack's API, the application TCP handle is 
	// immediately invalid after calling this function, so there 
	// is no longer any way to receive data from the TCP RX FIFO, 
	// even though the data is still there.  Leaving the data there 
	// could interfere with the remote node sending us a FIN if our
	// RX window is zero
	ps->rxTail = ps->rxHead;

	switch(ps->smState)
	{
		// These states are all already closed or don't need explicit disconnecting -- they will disconnect by themselves after a while
		//case TCP_CLOSED:
		//case TCP_LISTEN:
		//case TCP_CLOSING:
		//case TCP_TIME_WAIT:
		//case TCP_LOOPBACK_CLOSED
		//	return;

		// This state will close itself after some delay, however, 
		// this is handled so that the user can call TCPDisconnect() 
		// twice to immediately close a socket (using an RST) without 
		// having to get an ACK back from the remote node.  This is 
		// great for instance when the application determines that 
		// the remote node has been physically disconnected and 
		// already knows that no ACK will be returned.  Alternatively, 
		// if the application needs to immediately reuse the socket 
		// regardless of what the other node's state is in (half open).
		case TCP_FIN_WAIT_1:
		case TCP_FIN_WAIT_2:
		case TCP_LAST_ACK:
			SendTCP(hTCP, RST | ACK, FALSE);
			CloseSocket(hTCP);
			break;

		case TCP_SYN_SENT:
			CloseSocket(hTCP);
			break;

		case TCP_SYN_RECEIVED:
		case TCP_ESTABLISHED:
			// Send the FIN
			SendTCP(hTCP, FIN | ACK, TRUE);
			ps->smState = TCP_FIN_WAIT_1;
			break;

//		case TCP_CLOSE_WAIT:
//			// Send the FIN
//			SendTCP(hTCP, FIN | ACK, TRUE);
//			ps->smState = TCP_LAST_ACK;
//			break;
			
		case TCP_LOOPBACK:
			// Indicate to loopback owner that the loop is done
			// Loopback owner must clean up later
			ps->smState = TCP_LOOPBACK_CLOSED;
	}
}

/*********************************************************************
* Function:        void TCPFlush(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           s       - Socket whose data is to be transmitted.
*
* Output:          None
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
void TCPFlush(TCP_SOCKET hTCP)
{
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];
	SyncTCB(hTCP);

	if(ps->txHead != MyTCB.txUnackedTail)
	{
		// Send the TCP segment with all unacked bytes
		SendTCP(hTCP, PSH | ACK, TRUE);
	}
}



/*********************************************************************
* Function:        WORD TCPIsPutReady(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP: handle of socket to test
*
* Output:          Number of bytes that can be immediately placed 
*				   in the transmit buffer.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
WORD TCPIsPutReady(TCP_SOCKET hTCP)
{
	TCB_STUB	*ps;
	BYTE i;

	ps = &TCBStubs[hTCP];

	i = ps->smState;

	// Unconnected sockets shouldn't be transmitting anything.
//	if(!((i == TCP_ESTABLISHED) || (i == TCP_CLOSE_WAIT) ||
//	     (i == TCP_LOOPBACK_CLOSED) || (i == TCP_LOOPBACK)))
	if(!((i == TCP_ESTABLISHED) ||
	     (i == TCP_LOOPBACK_CLOSED) || (i == TCP_LOOPBACK)))
		return 0;

	// Calculate the free space in this socket's TX FIFO
	if(ps->txHead >= ps->txTail)
		return (ps->bufferRxStart - ps->bufferTxStart - 1) - (ps->txHead - ps->txTail);
	else
		return ps->txTail - ps->txHead - 1;
}


/*********************************************************************
* Function:        BOOL TCPPut(TCP_SOCKET hTCP, BYTE byte)
*
* PreCondition:    TCPIsPutReady(s) != 0
*
* Input:           hTCP: socket handle to use
*                  byte: a data byte to send
*
* Output:          TRUE if given byte was put in transmit buffer
*                  FALSE if transmit buffer is full.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
BOOL TCPPut(TCP_SOCKET hTCP, BYTE byte)
{
	TCB_STUB *ps;
	WORD wFreeTXSpace;

	ps = &TCBStubs[hTCP];

	wFreeTXSpace = TCPIsPutReady(hTCP);
	if(wFreeTXSpace == 0u)
		return FALSE;
	else if(wFreeTXSpace == 1u) // About to run out of space, lets transmit so the remote node might send an ACK back faster
		TCPFlush(hTCP);	

	// Send all current bytes if we are crossing half full
	// This is required to improve performance with the delayed 
	// acknowledgement algorithm
	if((!ps->Flags.bHalfFullFlush) && (wFreeTXSpace <= ((ps->bufferRxStart-ps->bufferTxStart)>>1)))
	{
		TCPFlush(hTCP);	
		ps->Flags.bHalfFullFlush = TRUE;
	}

	MACSetWritePtr(ps->txHead);
	MACPut(byte);
	if(++ps->txHead >= ps->bufferRxStart)
		ps->txHead = ps->bufferTxStart;

	// Send the last byte as a separate packet (likely will make the remote node send back ACK faster)
	if(wFreeTXSpace == 1u)
	{
		TCPFlush(hTCP);
	}
	// If not already enabled, start a timer so this data will 
	// eventually get sent even if the application doens't call
	// TCPFlush()
	else if(!ps->Flags.bTimer2Enabled)
	{
		ps->Flags.bTimer2Enabled = TRUE;
		ps->eventTime2 = (WORD)TickGetDiv256() + TCP_AUTO_TRANSMIT_TIMEOUT_VAL/256ull;
	}

	return TRUE;
}

/*********************************************************************
* Function:        WORD TCPPutArray(TCP_SOCKET hTCP, BYTE *data, WORD len)
*
* PreCondition:    None
*
* Input:           hTCP    - Socket handle to use
*                  data    - Pointer to data to put
*				   len     - Count of bytes to put
*
* Output:          Count of bytes actually placed in the TX buffer
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
WORD TCPPutArray(TCP_SOCKET hTCP, BYTE *data, WORD len)
{
	WORD wActualLen;
	WORD wFreeTXSpace;
	WORD wRightLen = 0;
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];

	wFreeTXSpace = TCPIsPutReady(hTCP);
	if(wFreeTXSpace == 0u)
	{
		TCPFlush(hTCP);
		return 0;
	}

	wActualLen = wFreeTXSpace;
	if(wFreeTXSpace > len)
		wActualLen = len;

	// Send all current bytes if we are crossing half full
	// This is required to improve performance with the delayed 
	// acknowledgement algorithm
	if((!ps->Flags.bHalfFullFlush) && (wFreeTXSpace <= ((ps->bufferRxStart-ps->bufferTxStart)>>1)))
	{
		TCPFlush(hTCP);	
		ps->Flags.bHalfFullFlush = TRUE;
	}
	
	// See if we need a two part put
	if(ps->txHead + wActualLen >= ps->bufferRxStart)
	{
		wRightLen = ps->bufferRxStart-ps->txHead;
		MACSetWritePtr(ps->txHead);
		MACPutArray(data, wRightLen);
		data += wRightLen;
		wActualLen -= wRightLen;
		ps->txHead = ps->bufferTxStart;
	}

	MACSetWritePtr(ps->txHead);
	MACPutArray(data, wActualLen);
	ps->txHead += wActualLen;

	// Send these bytes right now if we are out of TX buffer space
	if(wFreeTXSpace <= len)
	{
		TCPFlush(hTCP);
	}
	// If not already enabled, start a timer so this data will 
	// eventually get sent even if the application doens't call
	// TCPFlush()
	else if(!ps->Flags.bTimer2Enabled)
	{
		ps->Flags.bTimer2Enabled = TRUE;
		ps->eventTime2 = (WORD)TickGetDiv256() + TCP_AUTO_TRANSMIT_TIMEOUT_VAL/256ull;
	}

	return wActualLen + wRightLen;
}

#if defined(__18CXX)
WORD TCPPutROMArray(TCP_SOCKET hTCP, ROM BYTE *data, WORD len)
{
	WORD wActualLen;
	WORD wFreeTXSpace;
	WORD wRightLen = 0;
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];

	wFreeTXSpace = TCPIsPutReady(hTCP);
	if(wFreeTXSpace == 0u)
	{
		TCPFlush(hTCP);
		return 0;
	}

	// Send all current bytes if we are crossing half full
	// This is required to improve performance with the delayed 
	// acknowledgement algorithm
	if((!ps->Flags.bHalfFullFlush) && (wFreeTXSpace <= ((ps->bufferRxStart-ps->bufferTxStart)>>1)))
	{
		TCPFlush(hTCP);	
		ps->Flags.bHalfFullFlush = TRUE;
	}
	
	wActualLen = wFreeTXSpace;
	if(wFreeTXSpace > len)
		wActualLen = len;
	
	// See if we need a two part put
	if(ps->txHead + wActualLen >= ps->bufferRxStart)
	{
		wRightLen = ps->bufferRxStart-ps->txHead;
		MACSetWritePtr(ps->txHead);
		MACPutROMArray(data, wRightLen);
		data += wRightLen;
		wActualLen -= wRightLen;
		ps->txHead = ps->bufferTxStart;
	}

	MACSetWritePtr(ps->txHead);
	MACPutROMArray(data, wActualLen);
	ps->txHead += wActualLen;

	// Send these bytes right now if we are out of TX buffer space
	if(wFreeTXSpace <= len)
	{
		TCPFlush(hTCP);
	}
	// If not already enabled, start a timer so this data will 
	// eventually get sent even if the application doens't call
	// TCPFlush()
	else if(!ps->Flags.bTimer2Enabled)
	{
		ps->Flags.bTimer2Enabled = TRUE;
		ps->eventTime2 = (WORD)TickGetDiv256() + TCP_AUTO_TRANSMIT_TIMEOUT_VAL/256ull;
	}

	return wActualLen + wRightLen;
}
#endif

/*********************************************************************
* Function:        WORD TCPPutString(TCP_SOCKET hTCP, BYTE *data)
*
* PreCondition:    None
*
* Input:           hTCP    - Socket handle to use
*                  data    - Pointer to null terminated string to put
*
* Output:          BYTE* - *data pointer incremented by the number of 
*							bytes actually placed in the TX FIFO
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
BYTE* TCPPutString(TCP_SOCKET hTCP, BYTE *data)
{
	return data + TCPPutArray(hTCP, data, strlen((char*)data));
}

#if defined(__18CXX)
ROM BYTE* TCPPutROMString(TCP_SOCKET hTCP, ROM BYTE *data)
{
	return data + TCPPutROMArray(hTCP, data, strlenpgm((ROM char*)data));
}
#endif

/*********************************************************************
* Function:		WORD TCPGetTxFIFOFull(TCP_SOCKET hTCP)
*
* PreCondition:	TCPInit() is already called
*
* Input:		hTCP - TCP socket handle
*
* Output:		Number of bytes that are pending in TxFIFO.  
*				0 is returned if the FIFO is completely empty.
*
* Side Effects:	None
*
* Overview:		None
*
* Note:			None
********************************************************************/
WORD TCPGetTxFIFOFull(TCP_SOCKET hTCP)
{
	WORD wDataLen;
	WORD wFIFOSize;

	// Calculate total usable FIFO size
	wFIFOSize = TCBStubs[hTCP].bufferRxStart - TCBStubs[hTCP].bufferTxStart - 1;

	// Find out how many data bytes are free in the TX FIFO
	wDataLen = TCPIsPutReady(hTCP);

	return wFIFOSize - wDataLen;
}

/*********************************************************************
* Function:        void TCPDiscard(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP       - socket handle
*
* Output:          None
*
* Side Effects:    None
*
* Overview:        Removes all pending data from the socket's RX 
*				   FIFO.
*
* Note:            None
********************************************************************/
void TCPDiscard(TCP_SOCKET hTCP)
{
	TCB_STUB *ps;

	if(TCPIsGetReady(hTCP))
	{
		ps = &TCBStubs[hTCP];
	
		// Delete all data in the RX buffer
		ps->rxTail = ps->rxHead;
	
		// Send a Window update message to the remote node
		SendTCP(hTCP, ACK, TRUE);
	}
}


/*********************************************************************
* Function:        WORD TCPIsGetReady(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP       - socket to test
*
* Output:          Number of bytes that are available in socket 'hTCP' 
*				   for immediate retrieval
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
WORD TCPIsGetReady(TCP_SOCKET hTCP)
{
	TCB_STUB	*ps;

	ps = &TCBStubs[hTCP];
	
	// If the loopback is closed, reject all input
	if(ps->smState == TCP_LOOPBACK_CLOSED)
		return 0;

	if(ps->rxHead >= ps->rxTail)
		return ps->rxHead - ps->rxTail;
	else
		return (ps->bufferEnd - ps->rxTail + 1) + (ps->rxHead - ps->bufferRxStart);
}


/*********************************************************************
* Function:        BOOL TCPGet(TCP_SOCKET hTCP, BYTE *byte)
*
* PreCondition:    TCPInit() is already called
*
* Input:           hTCP    - socket
*                  byte    - Pointer to a byte.
*
* Output:          TRUE if a byte was read.
*                  FALSE if byte was not read.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
BOOL TCPGet(TCP_SOCKET hTCP, BYTE *byte)
{
	TCB_STUB	*ps;
	WORD GetReadyCount;

	GetReadyCount = TCPIsGetReady(hTCP);
	if(GetReadyCount == 0u)
		return FALSE;

	ps = &TCBStubs[hTCP];
	MACSetReadPtr(ps->rxTail);
	*byte = MACGet();
	if(++ps->rxTail > ps->bufferEnd)
		ps->rxTail = ps->bufferRxStart;

	// Send a window update if we've run out of data
	if(GetReadyCount == 1u)
	{
		ps->Flags.bTXASAP = 1;
	}
	// If not already enabled, start a timer so a window 
	// update will get sent to the remote node at some point
	else if(!ps->Flags.bTimer2Enabled)
	{
		ps->Flags.bTimer2Enabled = TRUE;
		ps->eventTime2 = (WORD)TickGetDiv256() + TCP_AUTO_TRANSMIT_TIMEOUT_VAL/256ull;
	}


	return TRUE;
}


/*********************************************************************
* Function:        WORD TCPGetArray(TCP_SOCKET hTCP, BYTE *buffer,
*                                      WORD len)
*
* PreCondition:    TCPInit() is already called
*
* Input:           hTCP    - socket handle
*                  buffer  - Buffer to hold received data.
*                  len     - Buffer length
*
* Output:          Number of bytes loaded into buffer.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
WORD TCPGetArray(TCP_SOCKET hTCP, BYTE *buffer, WORD len)
{
	WORD ActualLen;
	WORD RightLen = 0;
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];

	ActualLen = TCPIsGetReady(hTCP);
	if(len > ActualLen)
		len = ActualLen;

	// See if we need a two part get
	if(ps->rxTail + len > ps->bufferEnd)
	{
		RightLen = ps->bufferEnd - ps->rxTail + 1;
		MACSetReadPtr(ps->rxTail);
		MACGetArray(buffer, RightLen);
		if(buffer)
			buffer += RightLen;
		len -= RightLen;
		ps->rxTail = ps->bufferRxStart;
	}

	MACSetReadPtr(ps->rxTail);
	MACGetArray(buffer, len);
	ps->rxTail += len;

	// Send a window update if we've run low on data
	if(ActualLen - len <= len)
	{
		ps->Flags.bTXASAP = 1;
	}
	else if(!ps->Flags.bTimer2Enabled)
	// If not already enabled, start a timer so a window 
	// update will get sent to the remote node at some point
	{
		ps->Flags.bTimer2Enabled = TRUE;
		ps->eventTime2 = (WORD)TickGetDiv256() + TCP_AUTO_TRANSMIT_TIMEOUT_VAL/256ull;
	}

	return len + RightLen;
}


/*********************************************************************
* Function:		WORD TCPGetRxFIFOFree(TCP_SOCKET hTCP)
*
* PreCondition:	TCPInit() is already called
*
* Input:		hTCP - TCP socket handle
*
* Output:		Number of bytes that can still fit in the RX FIFO.  
*				0 is returned if the FIFO is completely full.
*
* Side Effects:	None
*
* Overview:		None
*
* Note:			None
********************************************************************/
WORD TCPGetRxFIFOFree(TCP_SOCKET hTCP)
{
	WORD wDataLen;
	WORD wFIFOSize;

	// Calculate total usable FIFO size
	wFIFOSize = TCBStubs[hTCP].bufferEnd - TCBStubs[hTCP].bufferRxStart;

	// Find out how many data bytes are actually in the RX FIFO
	wDataLen = TCPIsGetReady(hTCP);

	return wFIFOSize - wDataLen;
}

/*********************************************************************
* Function:		WORD TCPFindArrayEx(TCP_SOCKET hTCP, BYTE *cFindArray, WORD wLen, WORD wStart, WORD wEnd, BOOL bTextCompare)
*
* PreCondition:	TCPInit() is already called
*
* Input:		hTCP       - TCP socket handle
*				cFindArray - Pointer to an array of bytes to look for
*				wLen       - Length of cFindArray
*				wStart	   - Position within the RX FIFO to begin searching. 0 is the beginning of the FIFO.
*				wSearchLen - Number of bytes to look at from start position
*				bTextCompare- Search using a case-insensitive algorithm if TRUE, otherwise, do a binary 1 to 1 search
*
* Output:		0xFFFF: Array not found
*				0 to 65534: Location the array was found
*
* Side Effects:	None
*
* Overview:		None
*
* Note:			As an example, if the RX FIFO contains:
*				"PIC MCUs are good." 
*				and cFindArray -> "MCU"
*				and wLen -> 3
*				the value returned by the search function would be 4
********************************************************************/
WORD TCPFindArrayEx(TCP_SOCKET hTCP, BYTE *cFindArray, WORD wLen, WORD wStart, WORD wSearchLen, BOOL bTextCompare)
{
	WORD wDataLen;
	WORD wBytesUntilWrap;
	WORD wLocation;
	WORD wLenStart;
	BYTE *cFindArrayStart;
	BYTE i, j, k;
	BOOL isFinding;
	BYTE buffer[32];
	TCB_STUB *ps;

	if(wLen == 0u)
		return 0u;

	ps = &TCBStubs[hTCP];

	// Find out how many bytes are in the RX FIFO and return 
	// immediately if we won't possibly find a match
	wDataLen = TCPIsGetReady(hTCP) - wStart;
	if(wDataLen < wLen)
		return 0xFFFFu;
	if(wSearchLen && wDataLen > wSearchLen)
		wDataLen = wSearchLen;

	wLocation = ps->rxTail + wStart;
	if(wLocation > ps->bufferEnd)
		wLocation -= ps->bufferEnd - ps->bufferRxStart + 1;
	MACSetReadPtr(wLocation);
	wBytesUntilWrap = ps->bufferEnd - wLocation + 1;
	wLocation = wStart;
	wLenStart = wLen;
	cFindArrayStart = cFindArray;
	j = *cFindArray++;
	isFinding = FALSE;
	if(bTextCompare)
	{
		if(j >= 'a' && j <= 'z')
			j += 'A'-'a';
	}

	// Search for the array
	while(1)
	{
		// Figure out how big of a chunk to read
		k = sizeof(buffer);
		if(k > wBytesUntilWrap)
			k = wBytesUntilWrap;
		if((WORD)k > wDataLen)
			k = wDataLen;

		// Read a chunk of data into the buffer
		MACGetArray(buffer, (WORD)k);
		wBytesUntilWrap -= k;

		if(wBytesUntilWrap == 0)
		{
			MACSetReadPtr(ps->bufferRxStart);
			wBytesUntilWrap = 0xFFFFu;
		}

		// Convert everything to uppercase
		if(bTextCompare)
		{
			for(i = 0; i < k; i++)
			{
				if(buffer[i] >= 'a' && buffer[i] <= 'z')
					buffer[i] += 'A'-'a';

				if(j == buffer[i])
				{
					if(--wLen == 0)
						return wLocation-wLenStart + i + 1;
					j = *cFindArray++;
					isFinding = TRUE;
					if(j >= 'a' && j <= 'z')
						j += 'A'-'a';
				}
				else
				{
					wLen = wLenStart;
					if(isFinding)
					{
						cFindArray = cFindArrayStart;
						j = *cFindArray++;
						if(j >= 'a' && j <= 'z')
							j += 'A'-'a';
						isFinding = FALSE;
					}
				}
			}
		}
		else	// Compare as is
		{
			for(i = 0; i < k; i++)
			{
				if(j == buffer[i])
				{
					if(--wLen == 0)
						return wLocation-wLenStart + i + 1;
					j = *cFindArray++;
					isFinding = TRUE;
				}
				else
				{
					wLen = wLenStart;
					if(isFinding)
					{
						cFindArray = cFindArrayStart;
						j = *cFindArray++;
						isFinding = FALSE;
					}
				}
			}
		}

		// Check to see if it is impossible to find a match
		wDataLen -= k;
		if(wDataLen < wLen)
			return 0xFFFFu;

		wLocation += k;
	}
}

#if defined(__18CXX)
WORD TCPFindROMArrayEx(TCP_SOCKET hTCP, ROM BYTE *cFindArray, WORD wLen, WORD wStart, WORD wSearchLen, BOOL bTextCompare)
{
	WORD wDataLen;
	WORD wBytesUntilWrap;
	WORD wLocation;
	WORD wLenStart;
	ROM BYTE *cFindArrayStart;
	BYTE i, j, k;
	BOOL isFinding;
	BYTE buffer[32];
	TCB_STUB *ps;

	if(wLen == 0u)
		return 0u;

	ps = &TCBStubs[hTCP];

	// Find out how many bytes are in the RX FIFO and return 
	// immediately if we won't possibly find a match
	wDataLen = TCPIsGetReady(hTCP) - wStart;
	if(wDataLen < wLen)
		return 0xFFFFu;
	if(wSearchLen && wDataLen > wSearchLen)
		wDataLen = wSearchLen;

	wLocation = ps->rxTail + wStart;
	if(wLocation > ps->bufferEnd)
		wLocation -= ps->bufferEnd - ps->bufferRxStart + 1;
	MACSetReadPtr(wLocation);
	wBytesUntilWrap = ps->bufferEnd - wLocation + 1;
	wLocation = wStart;
	wLenStart = wLen;
	cFindArrayStart = cFindArray;
	j = *cFindArray++;
	isFinding = FALSE;
	if(bTextCompare)
	{
		if(j >= 'a' && j <= 'z')
			j += 'A'-'a';
	}

	// Search for the array
	while(1)
	{
		// Figure out how big of a chunk to read
		k = sizeof(buffer);
		if(k > wBytesUntilWrap)
			k = wBytesUntilWrap;
		if((WORD)k > wDataLen)
			k = wDataLen;

		// Read a chunk of data into the buffer
		MACGetArray(buffer, (WORD)k);
		wBytesUntilWrap -= k;

		if(wBytesUntilWrap == 0)
		{
			MACSetReadPtr(ps->bufferRxStart);
			wBytesUntilWrap = 0xFFFFu;
		}

		// Convert everything to uppercase
		if(bTextCompare)
		{
			for(i = 0; i < k; i++)
			{
				if(buffer[i] >= 'a' && buffer[i] <= 'z')
					buffer[i] += 'A'-'a';

				if(j == buffer[i])
				{
					if(--wLen == 0)
						return wLocation-wLenStart + i + 1;
					j = *cFindArray++;
					isFinding = TRUE;
					if(j >= 'a' && j <= 'z')
						j += 'A'-'a';
				}
				else
				{
					wLen = wLenStart;
					if(isFinding)
					{
						cFindArray = cFindArrayStart;
						j = *cFindArray++;
						if(j >= 'a' && j <= 'z')
							j += 'A'-'a';
						isFinding = FALSE;
					}
				}
			}
		}
		else	// Compare as is
		{
			for(i = 0; i < k; i++)
			{
				if(j == buffer[i])
				{
					if(--wLen == 0)
						return wLocation-wLenStart + i + 1;
					j = *cFindArray++;
					isFinding = TRUE;
				}
				else
				{
					wLen = wLenStart;
					if(isFinding)
					{
						cFindArray = cFindArrayStart;
						j = *cFindArray++;
						isFinding = FALSE;
					}
				}
			}
		}

		// Check to see if it is impossible to find a match
		wDataLen -= k;
		if(wDataLen < wLen)
			return 0xFFFFu;

		wLocation += k;
	}
}
#endif

WORD TCPFindEx(TCP_SOCKET hTCP, BYTE cFind, WORD wStart, WORD wSearchLen, BOOL bTextCompare)
{
	return TCPFindArrayEx(hTCP, &cFind, sizeof(cFind), wStart, wSearchLen, bTextCompare);
}


/*********************************************************************
* Function:        void TCPTick(void)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           None
*
* Output:          Each socket FSM is executed for any timeout
*                  situation.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
void TCPTick(void)
{
	TCP_SOCKET hTCP;
	TCB_STUB *ps;
	BOOL bRetransmit;
	BOOL bCloseSocket;
	BYTE vFlags;

	// Periodically all "not closed" sockets must perform timed operations
	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++)
	{
		ps = &TCBStubs[hTCP];

		vFlags = 0x00;
		bRetransmit = FALSE;
		bCloseSocket = FALSE;

		// Transmit ASAP data if the medium is available
		if(ps->Flags.bTXASAP)
		{
			if(MACIsTxReady())
				vFlags = ACK;
		}

		// Perform any needed window updates and data transmissions
		if(ps->Flags.bTimer2Enabled)
		{
			// See if the timeout has occured, and we need to send a new window update and pending data
			if((SHORT)(ps->eventTime2 - (WORD)TickGetDiv256()) <= (SHORT)0)
				vFlags = ACK;
		}

		// Process Delayed ACKnowledgement timer
		if(ps->Flags.bDelayedACKTimerEnabled)
		{
			// See if the timeout has occured and delayed ACK needs to be sent
			if((SHORT)(ps->delayedACKTime - (WORD)TickGetDiv256()) <= (SHORT)0)
				vFlags = ACK;
		}

		if(vFlags)
			SendTCP(hTCP, ACK, TRUE);

		// The TCP_CLOSED, TCP_LISTEN, and sometimes the TCP_ESTABLISHED 
		// state don't need any timeout events
		if(!ps->Flags.bTimerEnabled)
			continue;

		// If timeout has not occured, do not do anything.
		if((LONG)(TickGet() - ps->eventTime) < (LONG)0)
			continue;

		// Load up extended TCB information
		SyncTCB(hTCP);

		// A timeout has occured.  Respond to this timeout condition
		// depending on what state this socket is in.
		switch(ps->smState)
		{
			case TCP_SYN_SENT:
				// Keep sending SYN until we hear from remote node.
				// This may be for infinite time, in that case
				// caller must detect it and do something.
				vFlags = SYN;
				bRetransmit = TRUE;
				break;
	
			case TCP_SYN_RECEIVED:
				// We must receive ACK before timeout expires.
				// If not, resend SYN+ACK.
				// Abort, if maximum attempts counts are reached.
				if(MyTCB.retryCount < TCP_MAX_SYN_RETRIES)
				{
					vFlags = SYN | ACK;
					bRetransmit = TRUE;
				}
				else
				{
					if(ps->Flags.bServer)
					{
						vFlags = RST | ACK;
						bCloseSocket = TRUE;
					}
					else
					{
						vFlags = SYN;
					}
				}
				break;
	
			case TCP_ESTABLISHED:
				// Retransmit any unacknowledged data
				if(MyTCB.retryCount < TCP_MAX_RETRIES)
				{
					vFlags = PSH | ACK;
					bRetransmit = TRUE;
				}
				else
				{
					// No response back for too long, close connection
					// This could happen, for instance, if the communication 
					// medium was lost
					ps->smState = TCP_FIN_WAIT_1;
					vFlags = FIN | ACK;
				}
				break;
	
			case TCP_FIN_WAIT_1:
				if(MyTCB.retryCount < TCP_MAX_RETRIES)
				{
					// Send another FIN
					vFlags = FIN | ACK;
					bRetransmit = TRUE;
				}
				else
				{
					// Close on our own, we can't seem to communicate 
					// with the remote node anymore
					vFlags = RST | ACK;
					bCloseSocket = TRUE;
				}
				break;
	
			case TCP_FIN_WAIT_2:
				// Close on our own, we can't seem to communicate 
				// with the remote node anymore
				vFlags = RST | ACK;
				bCloseSocket = TRUE;
				break;

			case TCP_CLOSING:
				if(MyTCB.retryCount < TCP_MAX_RETRIES)
				{
					// Send another ACK+FIN (the FIN is retransmitted 
					// automatically since it hasn't been acknowledged by 
					// the remote node yet)
					vFlags = ACK;
					bRetransmit = TRUE;
				}
				else
				{
					// Close on our own, we can't seem to communicate 
					// with the remote node anymore
					vFlags = RST | ACK;
					bCloseSocket = TRUE;
				}
				break;
	
//			case TCP_TIME_WAIT:
//				// Wait around for a while (2MSL) and then goto closed state
//				bCloseSocket = TRUE;
//				break;
//			
//			case TCP_CLOSE_WAIT:
//				if(MyTCB.retryCount < TCP_MAX_RETRIES)
//				{
//					// Send ACK
//					vFlags = ACK;
//					bRetransmit = TRUE;
//				}
//				else
//				{
//					// We shouldn't get down here if the application properly closes the connection.
//					// However, if it doesn't, we still need to ensure socket reuse, so let's forcefully close the socket.
//					vFlags = RST | ACK;
//					bCloseSocket = TRUE;
//				}
//				break;
	
			case TCP_LAST_ACK:
				// Send some more FINs or close anyway
				if(MyTCB.retryCount < TCP_MAX_RETRIES)
				{
					vFlags = FIN | ACK;
					bRetransmit = TRUE;
				}
				else
				{
					vFlags = RST | ACK;
					bCloseSocket = TRUE;
				}
				break;
		}

		if(vFlags)
		{
			if(bRetransmit)
			{
				// Set the appropriate retry time
				MyTCB.retryCount++;
				MyTCB.retryInterval <<= 1;
		
				// Transmit all unacknowledged data over again
				// Roll back unacknowledged TX tail pointer to cause retransmit to occur
				MyTCB.MySEQ -= (LONG)(SHORT)(MyTCB.txUnackedTail - ps->txTail);
				if(MyTCB.txUnackedTail < ps->txTail)
					MyTCB.MySEQ -= (LONG)(SHORT)(ps->bufferRxStart - ps->bufferTxStart);
				MyTCB.txUnackedTail = ps->txTail;		
			}

			SendTCP(hTCP, vFlags, !bRetransmit);
		}
		
		if(bCloseSocket)
			CloseSocket(hTCP);
	}
}



/*********************************************************************
* Function:        BOOL TCPProcess(NODE_INFO* remote,
*                                  IP_ADDR *localIP,
*                                  WORD len)
*
* PreCondition:    TCPInit() is already called     AND
*                  TCP segment is ready in MAC buffer
*
* Input:           remote      - Remote node info
*                  len         - Total length of TCP semgent.
*
* Output:          TRUE if this function has completed its task
*                  FALSE otherwise
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
BOOL TCPProcess(NODE_INFO *remote, IP_ADDR *localIP, WORD len)
{
	TCP_HEADER      TCPHeader;
	PSEUDO_HEADER   pseudoHeader;
	TCP_SOCKET      socket;
	WORD_VAL        checksum1;
	WORD_VAL        checksum2;
	BYTE            optionsSize;

	// Calculate IP pseudoheader checksum.
	pseudoHeader.SourceAddress      = remote->IPAddr;
	pseudoHeader.DestAddress        = *localIP;
	pseudoHeader.Zero               = 0x0;
	pseudoHeader.Protocol           = IP_PROT_TCP;
	pseudoHeader.TCPLength          = len;

	SwapPseudoTCPHeader(pseudoHeader);

	checksum1.Val = ~CalcIPChecksum((BYTE*)&pseudoHeader,
		sizeof(pseudoHeader));

	// Now calculate TCP packet checksum in NIC RAM - should match
	// pesudo header checksum
	checksum2.Val = CalcIPBufferChecksum(len);

	// Compare checksums.
	if(checksum1.Val != checksum2.Val)
	{
		MACDiscardRx();
		return TRUE;
	}

#if defined(DEBUG_GENERATE_RX_LOSS)
	// Throw RX packets away randomly
	if(rand() > DEBUG_GENERATE_RX_LOSS)
	{
		MACDiscardRx();
		return TRUE;
	}
#endif

	// Retrieve TCP header.
	IPSetRxBuffer(0);
	MACGetArray((BYTE*)&TCPHeader, sizeof(TCPHeader));
	SwapTCPHeader(&TCPHeader);


	// Skip over options to retrieve data bytes
	optionsSize = (BYTE)((TCPHeader.DataOffset.Val << 2)-
		sizeof(TCPHeader));
	len = len - optionsSize - sizeof(TCPHeader);

	// Prevent loopback injection attacks.  Discard packets 
	// arriving externally and addressed from the loopback IP
	if(remote->IPAddr.Val == LOOPBACK_IP)
	{
		MACDiscardRx();
		return TRUE;
	}

	// Find matching socket.
	socket = FindMatchingSocket(&TCPHeader, remote);
	if(socket != INVALID_SOCKET)
	{
		HandleTCPSeg(socket, &TCPHeader, len);
	}
//	else
//	{
//		// \TODO: RFC 793 specifies that if the socket is closed 
//		// and a segment arrives, we should send back a RST if 
//		// the RST bit in the incoming packet is not set.  The 
//		// code to respond with the RST packet and believable SEQ 
//		// and ACK numbers needs to be implemented.
//		//if(!TCPHeader.Flags.bits.flagRST)
//		//	SendTCP(INVALID_SOCKET, RST, TRUE);
//	}

	// Finished with this packet, discard it and free the Ethernet RAM for new packets
	MACDiscardRx();

	return TRUE;
}


/*********************************************************************
* Function:        static void SendTCP(TCP_SOCKET hTCP, BYTE flags, BOOL bResetTimer)
*
* PreCondition:    TCPInit() is already called
*
* Input:           hTCP: Handle of socket to send packet with
*				   flags: Additional TCP flags to include
*				   bResetTimer: Indicates if this packet is a retransmission (no reset) or a new packet (reset required)
*
* Output:          A TCP segment is assembled and transmitted
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
static void SendTCP(TCP_SOCKET hTCP, BYTE flags, BOOL bResetTimer)
{
	TCB_STUB		*ps;
	WORD_VAL        wVal;
	TCP_HEADER      header;
	TCP_OPTIONS     options;
	PSEUDO_HEADER   pseudoHeader;
	WORD 			len;
	WORD			wEffectiveWindow;
	
	ps = &TCBStubs[hTCP];

	// Do not send anything on loopbacked ports
	if(ps->smState == TCP_LOOPBACK || ps->smState == TCP_LOOPBACK_CLOSED)
		return;

	SyncTCB(hTCP);

	// FINs must be handled specially
	if(flags & FIN)
	{
		ps->Flags.bTXFIN = 1;
		flags &= ~FIN;
	}

	// Status will now be synched, disable automatic future 
	// status transmissions
	ps->Flags.bTimer2Enabled = 0;
	ps->Flags.bDelayedACKTimerEnabled = 0;
	ps->Flags.bOneSegmentReceived = 0;
	ps->Flags.bTXASAP = 0;
	ps->Flags.bHalfFullFlush = 0;

	//  Make sure that we can write to the MAC transmit area
	while(!IPIsTxReady());

	// Put all socket application data in the TX space
	if(flags & (SYN | RST))
	{
		// Don't put any data in SYN and RST messages
		len = 0;
	}
	else
	{
		// Begin copying any application data over to the TX space
		if(ps->txHead == MyTCB.txUnackedTail)
		{
			// All caught up on data TX, no real data for this packet
			len = 0;

			// If we are to transmit a FIN, make sure we can put one in this packet
			if(ps->Flags.bTXFIN)
			{
				if(MyTCB.remoteWindow)
					flags |= FIN;
			}
		}
		else if(ps->txHead > MyTCB.txUnackedTail)
		{
			len = ps->txHead - MyTCB.txUnackedTail;
			wEffectiveWindow = MyTCB.remoteWindow;
			if(MyTCB.txUnackedTail >= ps->txTail)
				wEffectiveWindow -= MyTCB.txUnackedTail - ps->txTail;
			else
				wEffectiveWindow -= (ps->bufferRxStart - ps->bufferTxStart) - (ps->txTail - MyTCB.txUnackedTail);

			if(len > wEffectiveWindow)
				len = wEffectiveWindow;

			if(len > TCP_MAX_SEG_SIZE)
			{
				len = TCP_MAX_SEG_SIZE;
				ps->Flags.bTXASAP = 1;
			}

			// If we are to transmit a FIN, make sure we can put one in this packet
			if(ps->Flags.bTXFIN)
			{
				if((len != wEffectiveWindow) && (len != TCP_MAX_SEG_SIZE))
					flags |= FIN;
			}

			MACMemCopyAsync(BASE_TX_ADDR+sizeof(ETHER_HEADER)+sizeof(IP_HEADER)+sizeof(TCP_HEADER), MyTCB.txUnackedTail, len);
			MyTCB.txUnackedTail += len;
		}
		else
		{
			pseudoHeader.TCPLength = ps->bufferRxStart - MyTCB.txUnackedTail;
			len = pseudoHeader.TCPLength + ps->txHead - ps->bufferTxStart;

			wEffectiveWindow = MyTCB.remoteWindow;
			if(MyTCB.txUnackedTail >= ps->txTail)
				wEffectiveWindow -= MyTCB.txUnackedTail - ps->txTail;
			else
				wEffectiveWindow -= (ps->bufferRxStart - ps->bufferTxStart) - (ps->txTail - MyTCB.txUnackedTail);
				
			if(len > wEffectiveWindow)
				len = wEffectiveWindow;

			if(len > TCP_MAX_SEG_SIZE)
			{
				len = TCP_MAX_SEG_SIZE;
				ps->Flags.bTXASAP = 1;
			}

			// If we are to transmit a FIN, make sure we can put one in this packet
			if(ps->Flags.bTXFIN)
			{
				if((len != wEffectiveWindow) && (len != TCP_MAX_SEG_SIZE))
					flags |= FIN;
			}

			if(pseudoHeader.TCPLength > wEffectiveWindow)
				pseudoHeader.TCPLength = wEffectiveWindow;

			if(pseudoHeader.TCPLength > TCP_MAX_SEG_SIZE)
				pseudoHeader.TCPLength = TCP_MAX_SEG_SIZE;

			MACMemCopyAsync(BASE_TX_ADDR+sizeof(ETHER_HEADER)+sizeof(IP_HEADER)+sizeof(TCP_HEADER), MyTCB.txUnackedTail, pseudoHeader.TCPLength);
			pseudoHeader.TCPLength = len - pseudoHeader.TCPLength;
	
			// Copy any left over chunks of application data over
			if(pseudoHeader.TCPLength)
			{
				MACMemCopyAsync(BASE_TX_ADDR+sizeof(ETHER_HEADER)+sizeof(IP_HEADER)+sizeof(TCP_HEADER)+(ps->bufferRxStart-MyTCB.txUnackedTail), ps->bufferTxStart, pseudoHeader.TCPLength);
			}

			MyTCB.txUnackedTail += len;
			if(MyTCB.txUnackedTail >= ps->bufferRxStart)
				MyTCB.txUnackedTail -= ps->bufferRxStart-ps->bufferTxStart;
		}
	}

	// Ensure that all packets with data of some kind are 
	// retransmitted by TCPTick() until acknowledged
	// Pure ACK packets with no data are not ACKed back in TCP
	if(len || (flags & (SYN | FIN)))
	{
		if(bResetTimer)
		{
			MyTCB.retryCount = 0;
			MyTCB.retryInterval = TCP_START_TIMEOUT_VAL;
		}	

		ps->eventTime = TickGet() + MyTCB.retryInterval;
		ps->Flags.bTimerEnabled = 1;
	}

	header.SourcePort			= MyTCB.localPort.Val;
	header.DestPort				= MyTCB.remotePort.Val;
	header.SeqNumber			= MyTCB.MySEQ;
	header.AckNumber			= MyTCB.RemoteSEQ;
	header.Flags.bits.Reserved2	= 0;
	header.DataOffset.Reserved3	= 0;
	header.Flags.byte			= flags;
	header.UrgentPointer        = 0;

	// Update our send sequence number and ensure retransmissions 
	// of SYNs and FINs use the right sequence number
	MyTCB.MySEQ += (DWORD)len;
	if(flags & SYN)
	{
		if(MyTCB.flags.bSYNSent)
			header.SeqNumber--;
		else
		{
			MyTCB.MySEQ++;
			MyTCB.flags.bSYNSent = 1;
		}
	}
	if(flags & FIN)
	{
		if(MyTCB.flags.bFINSent)
			header.SeqNumber--;
		else
		{
			MyTCB.MySEQ++;
			MyTCB.flags.bFINSent = 1;
		}
	}

	// Calculate the amount of free space in the RX buffer area of this socket
	if(ps->rxHead >= ps->rxTail)
		header.Window = (ps->bufferEnd - ps->bufferRxStart) - (ps->rxHead - ps->rxTail);
	else
		header.Window = ps->rxTail - ps->rxHead - 1;

	// Calculate the amount of free space in the MAC RX buffer area and adjust window if needed
	wVal.Val = MACGetFreeRxSize()-64;
	if((SHORT)wVal.Val < (SHORT)0)
		wVal.Val = 0;
	// Force the remote node to throttle back if we are running low on general RX buffer space
	if(header.Window > wVal.Val)
		header.Window = wVal.Val;

	SwapTCPHeader(&header);


	len += sizeof(header);
	header.DataOffset.Val   = sizeof(header) >> 2;

	// Insert the MSS (Maximum Segment Size) TCP option if this is SYN packet
	if(flags & SYN)
	{
		len += sizeof(options);
		options.Kind = TCP_OPTIONS_MAX_SEG_SIZE;
		options.Length = 0x04;

		// Load MSS and swap to big endian
		options.MaxSegSize.Val = (((TCP_MAX_SEG_SIZE-4)&0x00FF)<<8) | (((TCP_MAX_SEG_SIZE-4)&0xFF00)>>8);

		header.DataOffset.Val   += sizeof(options) >> 2;
	}

	// Calculate IP pseudoheader checksum.
	pseudoHeader.SourceAddress	= AppConfig.MyIPAddr;
	pseudoHeader.DestAddress    = MyTCB.remote.IPAddr;
	pseudoHeader.Zero           = 0x0;
	pseudoHeader.Protocol       = IP_PROT_TCP;
	pseudoHeader.TCPLength		= len;
	SwapPseudoTCPHeader(pseudoHeader);
	header.Checksum = ~CalcIPChecksum((BYTE*)&pseudoHeader, sizeof(pseudoHeader));

	// Write IP header
	MACSetWritePtr(BASE_TX_ADDR + sizeof(ETHER_HEADER));
	IPPutHeader(&MyTCB.remote, IP_PROT_TCP, len);
	MACPutArray((BYTE*)&header, sizeof(header));
	if(flags & SYN)
		MACPutArray((BYTE*)&options, sizeof(options));

	// Wait for any previous copy operations to complete
	while(!MACIsMemCopyDone());

	// Update the TCP checksum
	MACSetReadPtr(BASE_TX_ADDR + sizeof(ETHER_HEADER) + sizeof(IP_HEADER));
	wVal.Val = CalcIPBufferChecksum(len);
#if defined(DEBUG_GENERATE_TX_LOSS)
	// Damage TCP checksums on TX packets randomly
	if(rand() > DEBUG_GENERATE_TX_LOSS)
	{
		wVal.Val++;
	}
#endif
	MACSetWritePtr(BASE_TX_ADDR + sizeof(ETHER_HEADER) + sizeof(IP_HEADER) + 16);
	MACPutArray((BYTE*)&wVal, sizeof(WORD));

	// Physically start the packet transmission over the network
	MACFlush();
}



/*********************************************************************
* Function:        static TCP_SOCKET FindMatchingSocket(TCP_HEADER *h,
*                                      NODE_INFO* remote)
*
* PreCondition:    TCPInit() is already called
*
* Input:           h           - TCP Header to be matched against.
*                  remote      - Node who sent this header.
*
* Output:          A socket that matches with given header and remote
*                  node is searched.
*                  If such socket is found, its index is returned
*                  else INVALID_SOCKET is returned.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
static TCP_SOCKET FindMatchingSocket(TCP_HEADER *h, NODE_INFO *remote)
{
	TCB_STUB *ps;
	TCP_SOCKET hTCP;
	TCP_SOCKET partialMatch;
	WORD hash;

	partialMatch = INVALID_SOCKET;
	hash = (remote->IPAddr.w[1]+remote->IPAddr.w[0] + h->SourcePort) ^ h->DestPort;

	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++ )
	{
		ps = &TCBStubs[hTCP];

		if(ps->smState == TCP_CLOSED)
		{
			continue;
		}
		else if(ps->smState == TCP_LISTEN)
		{
			if(ps->remoteHash.Val == h->DestPort)
				partialMatch = hTCP;
			continue;
		}
		else if(ps->remoteHash.Val != hash)
		{
			continue;
		}

		SyncTCB(hTCP);
		if(	h->DestPort == MyTCB.localPort.Val &&
			h->SourcePort == MyTCB.remotePort.Val &&
			remote->IPAddr.Val == MyTCB.remote.IPAddr.Val)
		{
			return hTCP;
		}
	}

	// We are not listening on this port, nor is a socket using it
	if(partialMatch == INVALID_SOCKET)
		return INVALID_SOCKET;

	// Set up the extended TCB with the info needed to establish a 
	// connection and return this socket to the caller
	ps = &TCBStubs[partialMatch];
	SyncTCB(partialMatch);

	ps->remoteHash.Val = hash;

	memcpy((void*)&MyTCB.remote, (void*)remote, sizeof(NODE_INFO));
	MyTCB.remotePort.Val = h->SourcePort;
	MyTCB.localPort.Val = h->DestPort;
	((DWORD_VAL*)(&MyTCB.MySEQ))->w[0] = rand();
	((DWORD_VAL*)(&MyTCB.MySEQ))->w[1] = rand();
	MyTCB.txUnackedTail	= ps->bufferTxStart;

	return partialMatch;
}



/*********************************************************************
* Function:        static void SwapTCPHeader(TCP_HEADER* header)
*
* PreCondition:    None
*
* Input:           header      - TCP Header to be swapped.
*
* Output:          Given header is swapped.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
static void SwapTCPHeader(TCP_HEADER* header)
{
	header->SourcePort      = swaps(header->SourcePort);
	header->DestPort        = swaps(header->DestPort);
	header->SeqNumber       = swapl(header->SeqNumber);
	header->AckNumber       = swapl(header->AckNumber);
	header->Window          = swaps(header->Window);
	header->Checksum        = swaps(header->Checksum);
	header->UrgentPointer   = swaps(header->UrgentPointer);
}



/*********************************************************************
* Function:        static void CloseSocket(TCP_SOCKET hTCP)
*
* PreCondition:    TCPInit() is already called
*
* Input:           hTCP  - Handle of socket to close/reinitialize
*
* Output:          Given socket information is reset and any
*                  buffered bytes held by this socket are discarded.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
static void CloseSocket(TCP_SOCKET hTCP)
{
	TCB_STUB *ps;
	
	ps = &TCBStubs[hTCP];
	SyncTCB(hTCP);

	ps->remoteHash.Val = MyTCB.localPort.Val;
	ps->txHead = ps->bufferTxStart;
	ps->txTail = ps->bufferTxStart;
	ps->rxHead = ps->bufferRxStart;
	ps->rxTail = ps->bufferRxStart;
	ps->smState = ps->Flags.bServer ? TCP_LISTEN : TCP_CLOSED;
	ps->Flags.bTimerEnabled = 0;
	ps->Flags.bTimer2Enabled = 0;
	ps->Flags.bDelayedACKTimerEnabled = 0;
	ps->Flags.bOneSegmentReceived = 0;
	ps->Flags.bHalfFullFlush = 0;
	ps->Flags.bTXASAP = 0;
	ps->Flags.bTXFIN = 0;
	ps->Flags.bSocketReset = 1;

	MyTCB.flags.bFINSent = 0;
	MyTCB.flags.bSYNSent = 0;
	MyTCB.txUnackedTail = ps->bufferTxStart;
	((DWORD_VAL*)(&MyTCB.MySEQ))->w[0] = rand();
	((DWORD_VAL*)(&MyTCB.MySEQ))->w[1] = rand();
	MyTCB.sHoleSize = -1;

}



/*********************************************************************
* Function:        static void HandleTCPSeg(TCP_SOCKET hTCP, TCP_HEADER *h, WORD len)
*
* PreCondition:    TCPInit() is already called     AND
*                  TCPProcess() is the caller.
*
* Input:           hTCP        - Socket that owns this segment
*                  h           - TCP Header
*                  len         - Total buffer length.
*
* Output:          TCP FSM is executed on given socket with
*                  given TCP segment.
*
* Side Effects:    None
*
* Overview:        None
*
* Note:            None
********************************************************************/
static void HandleTCPSeg(TCP_SOCKET hTCP, TCP_HEADER *h, WORD len)
{
	DWORD dwTemp;
	WORD wTemp;
	LONG lMissingBytes;
	WORD wMissingBytes;
	WORD wFreeSpace;
	TCB_STUB *ps;
	TCP_STATE localState;
	BYTE localHeaderFlags;
	DWORD localAckNumber;
	DWORD localSeqNumber;
	WORD wSegmentLength;
	BOOL bSegmentAcceptable;

	ps = &TCBStubs[hTCP];
	SyncTCB(hTCP);

	// Cache a few variables in local RAM.  
	// PIC18s take a fair amount of code and execution time to 
	// dereference pointers frequently.
	localState = ps->smState;
	localHeaderFlags = h->Flags.byte;
	localAckNumber = h->AckNumber;
	localSeqNumber = h->SeqNumber;

	// Handle TCP_LISTEN and TCP_SYN_SENT states
	// Both of these states will return, so code following this 
	// state machine need not check explicitly for these two 
	// states.
	switch(localState)
	{
		case TCP_LISTEN:
			// First: check RST flag
			if(localHeaderFlags & RST)
			{
				CloseSocket(hTCP);	// Unbind remote IP address/port info
				return;
			}

			// Second: check ACK flag, which would be invalid
			if(localHeaderFlags & ACK)
			{
				// Use a believable sequence number and reset the remote node
				MyTCB.MySEQ = localAckNumber;
				SendTCP(hTCP, RST, FALSE);
				CloseSocket(hTCP);	// Unbind remote IP address/port info
				return;
			}

			// Third: check for SYN flag, which is what we're looking for
			if(localHeaderFlags & SYN)
			{
				// We now have a sequence number for the remote node
				MyTCB.RemoteSEQ = localSeqNumber + 1;

				// Set Initial Send Sequence (ISS) number
				// Nothing to do on this step... ISS already set in CloseSocket()
				
				// Respond with SYN + ACK
				SendTCP(hTCP, SYN | ACK, TRUE);
				ps->smState = TCP_SYN_RECEIVED;
			}
			else
			{
				CloseSocket(hTCP);	// Unbind remote IP address/port info
			}

			// Fourth: check for other text and control
			// Nothing to do since we don't support this
			return;

		case TCP_SYN_SENT:
			// Second: check the RST bit
			// This is out of order because this stack has no API for 
			// notifying the application that the connection seems to 
			// be failing.  Instead, the application must time out and 
			// the stack will just keep trying in the mean time.
			if(localHeaderFlags & RST)
				return;

			// First: check ACK bit
			if(localHeaderFlags & ACK)
			{
				if(localAckNumber != MyTCB.MySEQ)
				{
					// Send a RST packet with SEQ = SEG.ACK, but retain our SEQ 
					// number for arivial of any other SYN+ACK packets
					localSeqNumber = MyTCB.MySEQ;	// Save our original SEQ number
					MyTCB.MySEQ = localAckNumber;	// Set SEQ = SEG.ACK
					SendTCP(hTCP, RST, TRUE);		// Send the RST
					MyTCB.MySEQ = localSeqNumber;	// Restore original SEQ number
					return;
				}
			}

			// Third: check the security and precedence
			// No such feature in this stack.  We want to accept all connections.

			// Fourth: check the SYN bit
			if(localHeaderFlags & SYN)
			{
				// We now have an initial sequence number and window size
				MyTCB.RemoteSEQ = localSeqNumber + 1;
				MyTCB.remoteWindow = h->Window;

				if(localHeaderFlags & ACK)
				{
					SendTCP(hTCP, ACK, TRUE);
					ps->smState = TCP_ESTABLISHED;
					ps->Flags.bTimerEnabled = 0;
				}
				else
				{
					SendTCP(hTCP, SYN | ACK, TRUE);
					ps->smState = TCP_SYN_RECEIVED;
				}
			}

			// Fifth: drop the segment if neither SYN or RST is set
			return;
	}

	//
	// First: check the sequence number
	//
	wSegmentLength = len;
	if(localHeaderFlags & FIN)
		wSegmentLength++;
	if(localHeaderFlags & SYN)
		wSegmentLength++;

	// Calculate the RX FIFO space
	if(ps->rxHead >= ps->rxTail)
		wFreeSpace = (ps->bufferEnd - ps->bufferRxStart) - (ps->rxHead - ps->rxTail);
	else
		wFreeSpace = ps->rxTail - ps->rxHead - 1;

	// Calculate the number of bytes ahead of our head pointer this segment skips
	lMissingBytes = localSeqNumber - MyTCB.RemoteSEQ;
	wMissingBytes = (WORD)lMissingBytes;
	
	// Run TCP acceptability tests to verify that this packet has a valid sequence number
	bSegmentAcceptable = FALSE;
	if(wSegmentLength)
	{
		// Check to see if we have free space, and if so, if any of the data falls within the freespace
		if(wFreeSpace)
		{
			// RCV.NXT =< SEG.SEQ < RCV.NXT+RCV.WND
			if((lMissingBytes >= (LONG)0) && (wFreeSpace > lMissingBytes))
				bSegmentAcceptable = TRUE;
			else
			{
				// RCV.NXT =< SEG.SEQ+SEG.LEN-1 < RCV.NXT+RCV.WND
				if((lMissingBytes + (LONG)wSegmentLength > (LONG)0) && (lMissingBytes <= (LONG)(SHORT)(wFreeSpace - wSegmentLength)))
					bSegmentAcceptable = TRUE;
			}
			
			if((lMissingBytes < (LONG)wFreeSpace) && ((SHORT)wMissingBytes + (SHORT)wSegmentLength > (SHORT)0))
				bSegmentAcceptable = TRUE;
		}
		// Segments with data are not acceptable if we have no free space
	}
	else
	{
		// Zero length packets are acceptable if they fall within our free space window
		// SEG.SEQ = RCV.NXT
		if(lMissingBytes == 0)
		{
			bSegmentAcceptable = TRUE;
		}
		else
		{
			// RCV.NXT =< SEG.SEQ < RCV.NXT+RCV.WND
			if((lMissingBytes >= (LONG)0) && (wFreeSpace > lMissingBytes))
				bSegmentAcceptable = TRUE;
		}
	}
	
	if(!bSegmentAcceptable)
	{
		// Unacceptable segment, drop it and respond appropriately
		if(!(localHeaderFlags & RST)) 
			SendTCP(hTCP, ACK, TRUE);
		return;
	}


	//
	// Second: check the RST bit
	//
	//
	// Fourth: check the SYN bit
	//
	// Note, that since the third step is not implemented, we can 
	// combine this second and fourth step into a single operation.
	if(localHeaderFlags & (RST | SYN))
	{
		CloseSocket(hTCP);
		return;
	}

	//
	// Third: check the security and precedence
	//
	// Feature not supported.  Let's process this segment.

	//
	// Fifth: check the ACK bit
	//
	if(!(localHeaderFlags & ACK))
		return;

	switch(localState)
	{
		case TCP_SYN_RECEIVED:
			if(localAckNumber != MyTCB.MySEQ)
			{
				// Send a RST packet with SEQ = SEG.ACK, but retain our SEQ 
				// number for arivial of any other correct packets
				localSeqNumber = MyTCB.MySEQ;	// Save our original SEQ number
				MyTCB.MySEQ = localAckNumber;	// Set SEQ = SEG.ACK
				SendTCP(hTCP, RST, TRUE);		// Send the RST
				MyTCB.MySEQ = localSeqNumber;	// Restore original SEQ number
				return;
			}
			ps->smState = TCP_ESTABLISHED;
			// No break

		case TCP_ESTABLISHED:
		case TCP_FIN_WAIT_1:
		case TCP_FIN_WAIT_2:
//		case TCP_CLOSE_WAIT:
		case TCP_CLOSING:
			// Drop the packet if it ACKs something we haven't sent
			if((LONG)(MyTCB.MySEQ - localAckNumber) < (LONG)0)
			{
				SendTCP(hTCP, ACK, TRUE);
				return;
			}

			// Throw away all ACKnowledged TX data:
			// Calculate what the last acknowledged sequence number was (ignoring any FINs we sent)
			dwTemp = MyTCB.MySEQ - (LONG)(SHORT)(MyTCB.txUnackedTail - ps->txTail);
			if(MyTCB.txUnackedTail < ps->txTail)
				dwTemp -= ps->bufferRxStart - ps->bufferTxStart;
	
			// Calcluate how many bytes were ACKed with this packet
			dwTemp = localAckNumber - dwTemp;
			if(((LONG)(dwTemp) > (LONG)0) && (dwTemp <= ps->bufferRxStart - ps->bufferTxStart))
			{
				ps->Flags.bHalfFullFlush = FALSE;
	
				// Bytes ACKed, free up the TX FIFO space
				ps->txTail += dwTemp;
				if(ps->txTail >= ps->bufferRxStart)
				{
					ps->txTail -= ps->bufferRxStart - ps->bufferTxStart;
				}
			}

			// No need to keep our retransmit timer going if we have nothing that needs ACKing anymore
			if(ps->txTail == ps->txHead)
			{
				// Make sure there isn't a "FIN byte in our TX FIFO"
				if(ps->Flags.bTXFIN == 0)
				{
					ps->Flags.bTimerEnabled = 0;
				}
				else
				{
					// "Throw away" FIN byte from our TX FIFO if it has been ACKed
					if(MyTCB.MySEQ == localAckNumber)
					{
						ps->Flags.bTimerEnabled = 0;
						ps->Flags.bTXFIN = 0;
						MyTCB.flags.bFINSent = 0;	
					}
				}
			}

			// Update the local stored copy of the RemoteWindow
			// If previously we had a zero window, and now we don't 
			// immediately send whatever was pending
			if((MyTCB.remoteWindow == 0u) && h->Window)
				ps->Flags.bTXASAP = 1;
			MyTCB.remoteWindow = h->Window;

			// A couple of states must do all of the TCP_ESTABLISHED stuff, but also a little more
			if(localState == TCP_FIN_WAIT_1)
			{
				// Check to see if our FIN has been ACKnowledged
				if(MyTCB.MySEQ == localAckNumber)
				{
					// Reset our timer for forced closure if the remote node 
					// doesn't send us a FIN in a timely manner.
					ps->eventTime = TickGet() + TCP_FIN_WAIT_2_TIMEOUT;
					ps->Flags.bTimerEnabled = 1;
					ps->smState = TCP_FIN_WAIT_2;
				}
			}
			else if(localState == TCP_FIN_WAIT_2)
			{
				// RFC noncompliance:
				// The remote node should not keep sending us data 
				// indefinitely after we send a FIN to it.  
				// However, some bad stacks may still keep sending 
				// us data indefinitely after ACKing our FIN.  To 
				// prevent this from locking up our socket, let's 
				// send a RST right now and close forcefully on 
				// our side.
				if(!(localHeaderFlags & FIN))
				{
					MyTCB.MySEQ = localAckNumber;	// Set SEQ = SEG.ACK
					SendTCP(hTCP, RST | ACK, FALSE);
					CloseSocket(hTCP);
					return;
				}
			}
			else if(localState == TCP_CLOSING)
			{
				// Check to see if our FIN has been ACKnowledged
				if(MyTCB.MySEQ == localAckNumber)
				{
					// RFC not recommended: We should be going to 
					// the TCP_TIME_WAIT state right here and 
					// starting a 2MSL timer, but since we have so 
					// few precious sockets, we can't afford to 
					// leave a socket waiting around doing nothing 
					// for a long time.  If the remote node does 
					// not recieve this ACK, it'll have to figure 
					// out on it's own that the connection is now 
					// closed.
					CloseSocket(hTCP);
				}

				return;
			}

			break;

		case TCP_LAST_ACK:
			// Check to see if our FIN has been ACKnowledged
			if(MyTCB.MySEQ == localAckNumber)
				CloseSocket(hTCP);
			return;

//		case TCP_TIME_WAIT:
//			// Nothing is supposed to arrive here.  If it does, reset the quiet timer.
//			SendTCP(hTCP, ACK, TRUE);
//			return;
	}

	//
	// Sixth: Check the URG bit
	//
	// Urgent packets are not supported in this stack, so we
	// will throw them away instead
	if(localHeaderFlags & URG)
		return;

	//
	// Seventh: Process the segment text
	//
	// Throw data away if in a state that doesn't accept data
//	if(localState == TCP_CLOSE_WAIT)
//		return;
	if(localState == TCP_CLOSING)
		return;
	if(localState == TCP_LAST_ACK)
		return;
//	if(localState == TCP_TIME_WAIT)
//		return;


	// See if there are bytes we must skip
	if((SHORT)wMissingBytes <= 0)
	{
		// Position packet read pointer to start of useful data area.
		IPSetRxBuffer((h->DataOffset.Val << 2) - wMissingBytes);
		len += wMissingBytes;		

		// Truncate packets that would overflow our TCP RX FIFO
		// and request a retransmit by sending a duplicate ACK
		if(len > wFreeSpace)
			len = wFreeSpace;

		MyTCB.RemoteSEQ += (DWORD)len;
	
		// Copy the application data from the packet into the socket RX FIFO
		// See if we need a two part copy (spans bufferEnd->bufferRxStart)
		if(ps->rxHead + len > ps->bufferEnd)
		{
			wTemp = ps->bufferEnd - ps->rxHead + 1;
			MACMemCopyAsync(ps->rxHead, -1, wTemp);
			MACMemCopyAsync(ps->bufferRxStart, -1, len - wTemp);
			ps->rxHead = ps->bufferRxStart + (len - wTemp);
		}
		else
		{
			MACMemCopyAsync(ps->rxHead, -1, len);
			ps->rxHead += len;
		}
	
		// See if we have a hole and other data waiting already in the RX FIFO
		if(MyTCB.sHoleSize != -1)
		{
			MyTCB.sHoleSize -= len;
			wTemp = MyTCB.wFutureDataSize + MyTCB.sHoleSize;
	
			// See if we just closed up a hole, and if so, advance head pointer
			if((SHORT)wTemp < (SHORT)0)
			{
				MyTCB.sHoleSize = -1;
			}
			else if(MyTCB.sHoleSize <= 0)
			{
				MyTCB.RemoteSEQ += wTemp;
				ps->rxHead += wTemp;
				if(ps->rxHead > ps->bufferEnd)
					ps->rxHead -= ps->bufferEnd - ps->bufferRxStart + 1;							
				MyTCB.sHoleSize = -1;
			}
		}
	} // This packet is out of order or we lost a packet, see if we can generate a hole to accomodate it
	else if((SHORT)wMissingBytes > 0)
	{
		// Truncate packets that would overflow our TCP RX FIFO
		if(len + wMissingBytes > wFreeSpace)
			len = wFreeSpace - wMissingBytes;
	
		// Position packet read pointer to start of useful data area.
		IPSetRxBuffer(h->DataOffset.Val << 2);

		// See if we need a two part copy (spans bufferEnd->bufferRxStart)
		if(ps->rxHead + wMissingBytes + len > ps->bufferEnd)
		{
			// Calculate number of data bytes to copy before wraparound
			wTemp = ps->bufferEnd - ps->rxHead + 1 - wMissingBytes;
			if((SHORT)wTemp >= 0)
			{
				MACMemCopyAsync(ps->rxHead + wMissingBytes, -1, wTemp);
				MACMemCopyAsync(ps->bufferRxStart, -1, len - wTemp);
			}
			else
			{
				MACMemCopyAsync(ps->rxHead + wMissingBytes - (ps->bufferEnd - ps->bufferRxStart + 1), -1, len);
			}
		}
		else
		{
			MACMemCopyAsync(ps->rxHead + wMissingBytes, -1, len);
		}
	
		// Record the hole is here
		if(MyTCB.sHoleSize == -1)
		{
			MyTCB.sHoleSize = wMissingBytes;
			MyTCB.wFutureDataSize = len;
		}
		else
		{
			// We already have a hole, see if we can shrink the hole 
			// or extend the future data size
			if(wMissingBytes < (WORD)MyTCB.sHoleSize)
			{
				if((wMissingBytes + len > (WORD)MyTCB.sHoleSize + MyTCB.wFutureDataSize) || (wMissingBytes + len < (WORD)MyTCB.sHoleSize))
					MyTCB.wFutureDataSize = len;
				else
					MyTCB.wFutureDataSize = (WORD)MyTCB.sHoleSize + MyTCB.wFutureDataSize - wMissingBytes;
				MyTCB.sHoleSize = wMissingBytes;
			}
			else if(wMissingBytes + len > (WORD)MyTCB.sHoleSize + MyTCB.wFutureDataSize)
			{
				// Make sure that there isn't a second hole between 
				// our future data and this TCP segment's future data
				if(wMissingBytes <= (WORD)MyTCB.sHoleSize + MyTCB.wFutureDataSize)
					MyTCB.wFutureDataSize += wMissingBytes + len - (WORD)MyTCB.sHoleSize - MyTCB.wFutureDataSize;
			}
			
		}
	
		// RFC Noncompliance:
		// Request an immediate fast retransmit from remote node 
		// by sending three duplicate ACKs.  This might improve 
		// performance, but then again, it might damage 
		// performance if the remote node invokes slow restart 
		// and/or collision avoidance algorithms.
		//SendTCP(hTCP, ACK, TRUE);
		//while(!MACIsTxReady());
		//MACFlush();
		//while(!MACIsTxReady());
		//MACFlush();
	}

	// Send back an ACK of the data (+SYN | FIN) we just received, 
	// if any.  To minimize bandwidth waste, we are implementing 
	// the delayed acknowledgement algorithm here, only sending 
	// back an immediate ACK if this is the second segment received.  
	// Otherwise, a 200ms timer will cause the ACK to be transmitted.
	if(wSegmentLength)
	{
		// For non-established sockets, let's delete all data in 
		// the RX buffer immediately after receiving it.  This is 
		// not really how TCP was intended to operate since a 
		// socket cannot receive any response after it sends a FIN,
		// but our TCP application API doesn't readily accomodate
		// receiving data after calling TCPDisconnect(), which 
		// invalidates the application TCP handle.  By deleting all 
		// data, we'll ensure that the RX window is nonzero and 
		// the remote node will be able to send us a FIN response, 
		// which needs an RX window of at least 1.
		if(localState != TCP_ESTABLISHED)
			ps->rxTail = ps->rxHead;

		if(ps->Flags.bOneSegmentReceived)
		{
			SendTCP(hTCP, ACK, TRUE);
			SyncTCB(hTCP);
			// bOneSegmentReceived is cleared in SendTCP(), so no need here
		}
		else
		{
			ps->Flags.bOneSegmentReceived = TRUE;	
		
			// Do not send an ACK immediately back.  Instead, we will 
			// perform delayed acknowledgements.  To do this, we will 
			// just start a timer
			if(!ps->Flags.bDelayedACKTimerEnabled)
			{
				ps->Flags.bDelayedACKTimerEnabled = TRUE;
				ps->delayedACKTime = (WORD)TickGetDiv256() + (WORD)((TCP_DELAYED_ACK_TIMEOUT)>>8);
			}
		}
	}

	// See if all data is acknowledged.  If so, there is no need 
	// to keep the retransmission timer running right now.
	if(ps->txHead == ps->txTail)
	{
		ps->Flags.bTimerEnabled = 0;
	}


	//
	// Eighth: check the FIN bit
	//
	if(localHeaderFlags & FIN)
	{
		// Note: Since we don't have a good means of storing "FIN bytes" 
		// in our TCP RX FIFO, we must ensure that FINs are processed 
		// in-order.
		if(MyTCB.RemoteSEQ + 1 == localSeqNumber + (DWORD)wSegmentLength)
		{
			// FINs are treated as one byte of data for ACK sequencing
			MyTCB.RemoteSEQ++;

			switch(localState)
			{
				case TCP_ESTABLISHED:
				case TCP_SYN_RECEIVED:
					// RFC in exact: Since we don't have an API that 
					// can allow the user to execute a CLOSE command, 
					// the stack will automatically perform the close 
					// operation immediately.  Thus, we are going to 
					// TCP_LAST_ACK immediately instead of 
					// TCP_CLOSE_WAIT
					ps->smState = TCP_LAST_ACK;
					SendTCP(hTCP, FIN | ACK, TRUE);
					//ps->smState = TCP_CLOSE_WAIT;
					return;
	
				case TCP_FIN_WAIT_1:
					if(MyTCB.MySEQ == localAckNumber)
					{
						// RFC not recommended: We should be going to 
						// the TCP_TIME_WAIT state right here and 
						// starting a 2MSL timer, but since we have so 
						// few precious sockets, we can't afford to 
						// leave a socket waiting around doing nothing 
						// for a long time.  If the remote node does 
						// not recieve this ACK, it'll have to figure 
						// out on it's own that the connection is now 
						// closed.
						SendTCP(hTCP, ACK, FALSE);
						CloseSocket(hTCP);
						return;
					}
					else
					{
						ps->smState = TCP_CLOSING;
					}
					break;
	
				case TCP_FIN_WAIT_2:
					// RFC not recommended: We should be going to 
					// the TCP_TIME_WAIT state right here and 
					// starting a 2MSL timer, but since we have so 
					// few precious sockets, we can't afford to 
					// leave a socket waiting around doing nothing 
					// for a long time.  If the remote node does 
					// not recieve this ACK, it'll have to figure 
					// out on it's own that the connection is now 
					// closed.
					SendTCP(hTCP, ACK, FALSE);
					CloseSocket(hTCP);
					return;
			}

			// Acknowledge receipt of FIN
			SendTCP(hTCP, ACK, TRUE);
		}
	}
}


/*********************************************************************
* Function:        BOOL TCPAdjustFIFOSize(TCP_SOCKET hTCP, WORD sRX, WORD sTX)
*
* PreCondition:    TCPInit() is already called.
*
* Input:           hTCP - TCP socket handle
*				   sRX	- Receive FIFO Size
*				   sTX	- Transmit FIFO Size
*
* Output:          TRUE: the FIFOs were adjusted
*				   FALSE: if the sRX and sTX were of improper size
*
* Side Effects:    Zeroes both buffers
*
* Overview:        Adjusts the RX and TX FIFO sizes for optimal speed
*
* Note:            sRX + sTX <= TCP_RX_FIFO_SIZE + TCP_TX_FIFO_SIZE
*				   If <, then TX buffer will get remainder of space.
********************************************************************/
BOOL TCPAdjustFIFOSize(TCP_SOCKET hTCP, WORD sRX, WORD sTX, BOOL destroyRX)
{
	TCB_STUB *ps;

	// Make sure space is available
	if(sRX + sTX > TCP_TX_FIFO_SIZE + TCP_RX_FIFO_SIZE)
		return FALSE;

	// Load up info on this socket
	ps = &TCBStubs[hTCP];
	SyncTCB(hTCP);

	// Move the RX buffer pointer - it's the one that divides the two
	ps->bufferRxStart = ps->bufferTxStart + sTX + 1;

	// Empty the TX buffer
	MyTCB.txUnackedTail = ps->bufferTxStart;
	ps->txTail = ps->bufferTxStart;
	ps->txHead = ps->bufferTxStart;
	
	// Empty the RX buffer
	if(destroyRX || ps->rxTail < ps->bufferRxStart || ps->rxHead < ps->bufferRxStart)
	{	
		ps->rxTail = ps->bufferRxStart;
		ps->rxHead = ps->bufferRxStart;
	}

	// Send a window update to notify remote node of change
	if(ps->smState == TCP_ESTABLISHED)
		SendTCP(hTCP, ACK, TRUE);

	return TRUE;

}

/*********************************************************************
 * Function:        TCP_SOCKET TCPOpenLoopback(WORD destPort)
 *
 * PreCondition:    TCPInit() is already called
 *
 * Input:           destPort - the desired listening port
 *
 * Output:          A socket that matches with given header and remote
 *                  node is searched.
 *                  If such socket is found, its index is returned
 *                  else INVALID_SOCKET is returned.
 *
 * Side Effects:    None
 *
 * Overview:        None
 *
 * Note:            Only for incoming connections now, could be
 *					expanded to support outgoing connections.
********************************************************************/
TCP_SOCKET TCPOpenLoopback(WORD destPort)
{
	TCB_STUB *ps;
	TCP_SOCKET hTCP;

	//find a socket listening on the right port
	for(hTCP = 0; hTCP < MAX_TCP_SOCKETS; hTCP++ )
	{
		ps = &TCBStubs[hTCP];
		if(ps->smState == TCP_LISTEN && ps->remoteHash.Val == destPort)
		{
			//lock the socket
			ps->smState = TCP_LOOPBACK;
			
			//reset all the buffer pointers
			ps->txHead = ps->bufferTxStart;
			ps->txTail = ps->bufferTxStart;
			ps->rxHead = ps->bufferRxStart;
			ps->rxTail = ps->bufferRxStart;

			//return the socket handle
			return hTCP;
		}
	}

	//if no socket was found, then return a flag
	return INVALID_SOCKET;
}

/*********************************************************************
 * Function:        BOOL TCPCloseLoopback(TCP_SOCKET hTCP)
 *
 * PreCondition:    TCPInit() is already called
 *
 * Input:           hTCP - the TCP loopback handle to close
 *
 * Output:          TRUE if the loopback was closed
 *					FALSE if the socket was not a loopback
 *
 * Side Effects:    None
 *
 * Overview:        None
 *
 * Note:            Only for incoming connections now, could be
 *					expanded to support outgoing connections.
********************************************************************/
BOOL TCPCloseLoopback(TCP_SOCKET hTCP)
{
	if(TCBStubs[hTCP].smState != TCP_LOOPBACK && 
			TCBStubs[hTCP].smState != TCP_LOOPBACK_CLOSED)
		return FALSE;
	
	CloseSocket(hTCP);
	
	return TRUE;
	
}

/*********************************************************************
 * Function:        BOOL TCPIsLoopback(TCP_SOCKET hTCP)
 *
 * PreCondition:    TCPInit() is already called
 *
 * Input:           hTCP - the TCP loopback handle to close
 *
 * Output:          TRUE if the socket is a loopback
 *					FALSE if the socket is not a loopback
 *
 * Side Effects:    None
 *
 * Overview:        None
 *
 * Note:            None
********************************************************************/
BOOL TCPIsLoopback(TCP_SOCKET hTCP)
{
	if(TCBStubs[hTCP].smState != TCP_LOOPBACK && 
			TCBStubs[hTCP].smState != TCP_LOOPBACK_CLOSED)
		return FALSE;
	
	return TRUE;
	
}

/*********************************************************************
 * Function:        WORD TCPInject(TCP_SOCKET hTCP, BYTE *data, 
 *									WORD len)
 *
 * PreCondition:    TCPInit() is already called.
 *
 * Input:           hTCP    - Socket handle to use
 *                  data    - Pointer to data to inject
 *				   len     - Number of bytes in data
 *
 * Output:          Count of bytes actually placed in the RX buffer
 *
 * Side Effects:    None
 *
 * Overview:        Injects data into a socket's RX FIFO
 *
 * Note:            None
********************************************************************/
WORD TCPInject(TCP_SOCKET hTCP, BYTE *buffer, WORD len)
{
	WORD wTemp;
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];
	
	//if the connection is finished, pretend to accept bytes
	if(ps->smState == TCP_LOOPBACK_CLOSED)
		return len;
	
	//make sure there's enough space
	wTemp = TCPGetRxFIFOFree(hTCP);
	if(wTemp < len)
		len = wTemp;
	
	// Copy the data from the buffer into the RX FIFO
	// See if we need a two part copy (spans bufferEnd->bufferRxStart)
	if(ps->rxHead + len > ps->bufferEnd)
	{
		wTemp = ps->bufferEnd - ps->rxHead + 1;
		MACSetWritePtr(ps->rxHead);
		MACPutArray(buffer, wTemp);
		buffer += wTemp;
		MACSetWritePtr(ps->bufferRxStart);
		MACPutArray(buffer, len - wTemp);
		ps->rxHead = ps->bufferRxStart + (len - wTemp);
	}
	else
	{
		MACSetWritePtr(ps->rxHead);
		MACPutArray(buffer, len);
		ps->rxHead += len;
	}
	
	return len;
	
}

/*********************************************************************
 * Function:        WORD TCPSteal(TCP_SOCKET hTCP, BYTE *data, 
 *									WORD len)
 *
 * PreCondition:    TCPInit() is already called.
 *
 * Input:           hTCP    - Socket handle to use
 *                  data    - Pointer to a buffer to fill
 *				   len     - Number of bytes to read
 *
 * Output:          Count of bytes actually placed in *data
 *
 * Side Effects:    None
 *
 * Overview:        Removes data from a socket's TX FIFO and places
 *					it in *data 
 *
 * Note:            None
********************************************************************/
WORD TCPSteal(TCP_SOCKET hTCP, BYTE *buffer, WORD len)
{
	WORD ActualLen;
	WORD RightLen = 0;
	TCB_STUB *ps;

	ps = &TCBStubs[hTCP];

	ActualLen = TCPGetTxFIFOFull(hTCP);
	if(len > ActualLen)
		len = ActualLen;

	// See if we need a two part get	
	if(ps->txTail + len >= ps->bufferRxStart)
	{
		RightLen = ps->bufferRxStart - ps->txTail;
		MACSetReadPtr(ps->txTail);
		MACGetArray(buffer, RightLen);
		if(buffer)
			buffer += RightLen;
		len -= RightLen;
		ps->txTail = ps->bufferTxStart;
	}

	MACSetReadPtr(ps->txTail);
	MACGetArray(buffer, len);
	ps->txTail += len;
	
	return len + RightLen;
	
}

#endif //#if defined(STACK_USE_TCP)
